﻿using SimpleVectorMatrixPlayground.VectorMatrix;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SimpleVectorMatrixPlayground.Tests
{
    class Mat2x2Tests
    {
        private readonly ITestHandler testHandler;
        private readonly AssertionUtil Assert;

        public Mat2x2Tests(ITestHandler ts)
        {
            this.testHandler = ts;
            Assert = new AssertionUtil(ts);
        }

        public void Run()
        {
            testHandler.TestSuiteStart("Matrix2x2 Tests");

            RunCreationTests();
            RunElementAccessTests();
            RunRowColumnAccessTests();

            RunAdditionSubtractionTests();
            RunMultiplyScalarTests();
            RunMultiplicationTests();

            RunTransposeTests();
            RunNegateTests();

            RunDeterminantTests();
            RunInverseTests();

            RunRotationTests();
            RunScaleShearingTests();

            RunEqualTests();

            testHandler.TestSuiteEnd();
        }

        private void RunCreationTests()
        {
            testHandler.TestGroupStart("Creation");

            Assert.AssertEqual(Matrix2x2.Create(), 0, 0, 0, 0);
            Assert.AssertEqual(Matrix2x2.CreateIdentity(), 1, 0, 0, 1);
            Assert.AssertEqual(Matrix2x2.Create(1, -2, 3.4, 4), 1, -2, 3.4, 4);
            Assert.AssertEqual(Matrix2x2.Create(new double[2, 2] { { 11, 2 }, { 3, 4 } }), 11, 2, 3, 4);

            testHandler.TestGroupEnd();
        }

        private void RunElementAccessTests()
        {
            testHandler.TestGroupStart("Element access");

            Matrix2x2 mat1 = Matrix2x2.Create(-1, 11.1, -2.5, 7.1);
            Assert.AssertEqual(mat1[0, 0], -1);
            Assert.AssertEqual(mat1[0, 1], 11.1);
            Assert.AssertEqual(mat1[1, 0], -2.5);
            Assert.AssertEqual(mat1[1, 1], 7.1);
            Assert.AssertException<ArgumentOutOfRangeException>(() => mat1[2, 1] = 0);
            Assert.AssertException<ArgumentOutOfRangeException>(() => mat1[1, 2] = 0);

            Matrix2x2 mat2 = Matrix2x2.Create(-1, -25, -4.5, 12);
            Assert.AssertEqual(mat2[0], -1);
            Assert.AssertEqual(mat2[1], -25);
            Assert.AssertEqual(mat2[2], -4.5);
            Assert.AssertEqual(mat2[3], 12);
            Assert.AssertException<ArgumentOutOfRangeException>(() => mat2[4] = 0);

            testHandler.TestGroupEnd();
        }

        private void RunRowColumnAccessTests()
        {
            testHandler.TestGroupStart("Row / Column access");

            Matrix2x2 mat1 = Matrix2x2.Create(-1, 2, 5, -4);
            Assert.AssertEqual(mat1.GetRow(0)[0], -1);
            Assert.AssertEqual(mat1.GetRow(0)[1], 2);
            Assert.AssertEqual(mat1.GetRow(1)[0], 5);
            Assert.AssertEqual(mat1.GetRow(1)[1], -4);
            Assert.AssertException<ArgumentOutOfRangeException>( () => mat1.GetRow(2));
            Assert.AssertEqual(mat1.GetRowVector(0).X, -1);
            Assert.AssertEqual(mat1.GetRowVector(0).Y, 2);
            Assert.AssertEqual(mat1.GetRowVector(1).X, 5);
            Assert.AssertEqual(mat1.GetRowVector(1).Y, -4);
            Assert.AssertException<ArgumentOutOfRangeException>(() => mat1.GetRowVector(2));

            Assert.AssertEqual(mat1.GetColumn(0)[0], -1);
            Assert.AssertEqual(mat1.GetColumn(0)[1], 5);
            Assert.AssertEqual(mat1.GetColumn(1)[0], 2);
            Assert.AssertEqual(mat1.GetColumn(1)[1], -4);
            Assert.AssertException<ArgumentOutOfRangeException>(() => mat1.GetColumn(2));
            Assert.AssertEqual(mat1.GetColumnVector(0).X, -1);
            Assert.AssertEqual(mat1.GetColumnVector(0).Y, 5);
            Assert.AssertEqual(mat1.GetColumnVector(1).X, 2);
            Assert.AssertEqual(mat1.GetColumnVector(1).Y, -4);
            Assert.AssertException<ArgumentOutOfRangeException>(() => mat1.GetColumnVector(2));

            testHandler.TestGroupEnd();
        }

        private void RunAdditionSubtractionTests()
        {
            testHandler.TestGroupStart("Addition / Subtraction");

            Assert.AssertEqual(Matrix2x2.CreateIdentity() + Matrix2x2.CreateIdentity(), 2, 0, 0, 2);
            Assert.AssertEqual(Matrix2x2.Create(3, 2, 0, 1) + Matrix2x2.Create(1, 3, 2, 0), 4, 5, 2, 1);
            Assert.AssertEqual(Matrix2x2.Create(1, 2, 3, 4) + Matrix2x2.Create(-1, -2, -3, -4), Matrix2x2.Create());

            Assert.AssertEqual(Matrix2x2.Create(2, 0, 0, 2) - Matrix2x2.CreateIdentity(), Matrix2x2.CreateIdentity());
            Assert.AssertEqual(Matrix2x2.Create(4, 5, 2, 1) - Matrix2x2.Create(3, 2, 0, 1), Matrix2x2.Create(1, 3, 2, 0));

            testHandler.TestGroupEnd();
        }

        private void RunMultiplyScalarTests()
        {
            testHandler.TestGroupStart("Multiplication (scalar)");

            Assert.AssertEqual(Matrix2x2.CreateIdentity() * 12, 12, 0, 0, 12);
            Assert.AssertEqual(Matrix2x2.Create(3, 2, 0, 1) * -2, -6, -4, 0, -2);

            testHandler.TestGroupEnd();
        }

        private void RunMultiplicationTests()
        {
            testHandler.TestGroupStart("Multiplication");

            Assert.AssertEqual(Matrix2x2.CreateIdentity() * Matrix2x2.CreateIdentity(), Matrix2x2.CreateIdentity());
            Assert.AssertEqual(Matrix2x2.Create(2, 5, 1, 0) * Matrix2x2.Create(3, 4, 1, 1), Matrix2x2.Create(11, 13, 3, 4));
            Assert.AssertEqual(Matrix2x2.Create(0.5, 5, -2, -13) * Matrix2x2.Create(-4, 7, 5, 23), Matrix2x2.Create(-2 + 25, 3.5 + 115, 8 - 65, -14 - 299));

            testHandler.TestGroupEnd();
        }

        private void RunTransposeTests()
        {
            testHandler.TestGroupStart("Transpose");

            Assert.AssertEqual(Matrix2x2.CreateIdentity().Transpose(), Matrix2x2.CreateIdentity());
            Assert.AssertEqual(Matrix2x2.Create(1, 2, 3, 4).Transpose(), Matrix2x2.Create(1, 3, 2, 4));

            testHandler.TestGroupEnd();
        }

        private void RunNegateTests()
        {
            testHandler.TestGroupStart("Negation");

            Assert.AssertEqual(Matrix2x2.CreateIdentity().Negate(), Matrix2x2.Create(-1, 0, 0, -1));
            Assert.AssertEqual(Matrix2x2.Create(1, 2, 3, 4).Negate(), Matrix2x2.Create(-1, -2, -3 , -4));

            testHandler.TestGroupEnd();
        }

        private void RunDeterminantTests()
        {
            testHandler.TestGroupStart("Determinant");

            Assert.AssertEqual(Matrix2x2.CreateIdentity().Determinant(), 1);
            Assert.AssertEqual(Matrix2x2.Create().Determinant(), 0);
            Assert.AssertEqual(Matrix2x2.Create(-3, 2, 7, 5).Determinant(), -29);

            Assert.AssertEqual(Matrix2x2.Create(4, 0, 0, 6).Determinant(), 24);
            Assert.AssertEqual(Matrix2x2.Create(0, -4, 6, 0).Determinant(), 24);
            Assert.AssertEqual(Matrix2x2.Create(0, 4, 6, 0).Determinant(), -24);

            // the inverse of a matrix has determinant 1 / det
            Assert.AssertEqual(Matrix2x2.Create(4, 0, 0, 6).Inverse().Determinant(), 1.0 / 24, 0.000001);
            Assert.AssertEqual(Matrix2x2.Create(0, -4, 6, 0).Inverse().Determinant(), 1.0 / 24, 0.000001);
            Assert.AssertEqual(Matrix2x2.Create(0, 4, 6, 0).Inverse().Determinant(), -1.0 / 24, 0.000001);

            // test that the determinant of 2x2 matrix is actually the same as the perp dot product of its vectors
            Matrix2x2 mat1 = Matrix2x2.Create(20, 30, 45, 70);
            Assert.AssertEqual(mat1.Determinant(), 50);
            Assert.AssertEqual(Vector2.PerpDotProduct(mat1.GetRowVector(0), mat1.GetRowVector(1)), 50);
            Assert.AssertEqual(Vector2.PerpDotProduct(mat1.GetColumnVector(0), mat1.GetColumnVector(1)), 50);

            Assert.AssertEqual(mat1.Inverse().Determinant(), 1.0 / 50, 0.000001);

            testHandler.TestGroupEnd();
        }

        private void RunInverseTests()
        {
            testHandler.TestGroupStart("Inverse");

            Assert.AssertEqual(Matrix2x2.CreateIdentity().Inverse(), Matrix2x2.Create(1, 0, 0, 1));
            Assert.AssertEqual(Matrix2x2.CreateIdentity() * Matrix2x2.Create(1, 0, 0, 1), Matrix2x2.CreateIdentity());

            Assert.AssertEqual(Matrix2x2.Create(1, 2, -2, 1).Inverse(), Matrix2x2.Create(0.2, -0.4, 0.4, 0.2));
            Assert.AssertEqual(Matrix2x2.Create(1, 2, -2, 1) * Matrix2x2.Create(0.2, -0.4, 0.4, 0.2), Matrix2x2.CreateIdentity());

            testHandler.TestGroupEnd();
        }

        public void RunRotationTests()
        {
            testHandler.TestGroupStart("Rotation");
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(0) * Vector2.Create(-8, 4), Vector2.Create(-8, 4), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI * 2) * Vector2.Create(12, -2), Vector2.Create(12, -2), 0.0001);

            // counter clockwise rotation
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 2) * Vector2.Create(1, 0), Vector2.Create(0, 1), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 2) * Vector2.Create(1, 0), Vector2.Create(0, 1), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 2) * Vector2.Create(0, 5), Vector2.Create(-5, 0), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 2) * Vector2.Create(-2.5, 0), Vector2.Create(0, -2.5), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 2) * Vector2.Create(0, -88), Vector2.Create(88, 0), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 4) * Vector2.Create(1, 1), Vector2.Create(0, Math.Sqrt(2)), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(-Math.PI / 4) * Vector2.Create(1, 1), Vector2.Create(Math.Sqrt(2), 0), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 4) * Vector2.Create(4, 4), Vector2.Create(0, Math.Sqrt(32)), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(-Math.PI / 2) * Vector2.Create(-4, 3), Vector2.Create(3, 4), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(-Math.PI) * Vector2.Create(-4, 3), Vector2.Create(4, -3), 0.0001);

            // clockwise rotation
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(-Math.PI / 2) * Vector2.Create(1, 0), Vector2.Create(0, -1), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrixCW(Math.PI / 2) * Vector2.Create(1, 0), Vector2.Create(0, -1), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrixCW(Math.PI / 4) * Vector2.Create(4, 4), Vector2.Create(Math.Sqrt(32), 0), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrixCW(Math.PI / 2) * Vector2.Create(0, -3), Vector2.Create(-3, 0), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrixCW(Math.PI / 2) * Vector2.Create(-4, 0), Vector2.Create(0, 4), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrixCW(Math.PI / 2) * Vector2.Create(0, 5), Vector2.Create(5, 0), 0.0001);

            // same rotation counterclockwise and clockwise 
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 6) * Vector2.Create(-7, 8),
                        Matrix2x2.GetRotationMatrixCW(2 * Math.PI - Math.PI / 6) * Vector2.Create(-7, 8), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(2 * Math.PI - Math.PI / 7) * Vector2.Create(56, -9.1),
                        Matrix2x2.GetRotationMatrixCW(Math.PI / 7) * Vector2.Create(56, -9.1), 0.0001);

            // multiply a "column" vector to the right side of a matrix is the same as multiplying the same vector
            // treated as "row" vector to the left of the transposed matrix
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 2) * Vector2.Create(0, 2.5),
                        Vector2.Create(0, 2.5) * (Matrix2x2.GetRotationMatrix(Math.PI / 2).Transpose()), 0.0001);
            Assert.AssertEqual(Matrix2x2.GetRotationMatrix(Math.PI / 3) * Vector2.Create(-1.1, -7.9),
                        Vector2.Create(-1.1, -7.9) * (Matrix2x2.GetRotationMatrix(Math.PI / 3).Transpose()), 0.0001);

            testHandler.TestGroupEnd();
        }

        private void RunScaleShearingTests()
        {
            testHandler.TestGroupStart("Scaling / Shearing");

            Assert.AssertEqual(Matrix2x2.GetScalingMatrix(3, 3) * Vector2.Create(), Vector2.Create(), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetScalingMatrix(3, 3) * Vector2.Create(1, 1), Vector2.Create(3, 3), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetScalingMatrix(4.5, 1.2) * Vector2.Create(4, 2), Vector2.Create(18, 2.4), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetScalingMatrix(-3, -1) * Vector2.Create(4, 2), Vector2.Create(-12, -2), 0.00001);

            Assert.AssertEqual(Matrix2x2.GetShearingMatrixX(2) * Vector2.Create(), Vector2.Create(), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixX(2) * Vector2.Create(1, 0), Vector2.Create(1, 0), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixX(4) * Vector2.Create(1, 1), Vector2.Create(5, 1), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixX(3.4) * Vector2.Create(0, 1), Vector2.Create(3.4, 1), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixX(-2) * Vector2.Create(0, 1), Vector2.Create(-2, 1), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixX(3.4) * Vector2.Create(0, -2), Vector2.Create(-6.8, -2), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixX(-3.4) * Vector2.Create(0, -2), Vector2.Create(6.8, -2), 0.00001);

            Assert.AssertEqual(Matrix2x2.GetShearingMatrixY(2) * Vector2.Create(), Vector2.Create(), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixY(2) * Vector2.Create(1, 0), Vector2.Create(1, 2), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixY(2) * Vector2.Create(0, 1), Vector2.Create(0, 1), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixY(2) * Vector2.Create(1, 1), Vector2.Create(1, 3), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixY(4) * Vector2.Create(-1, 2), Vector2.Create(-1, -2), 0.00001);
            Assert.AssertEqual(Matrix2x2.GetShearingMatrixY(-4) * Vector2.Create(-1, 2), Vector2.Create(-1, 6), 0.00001);

            testHandler.TestGroupEnd();
        }

        private void RunEqualTests()
        {
            testHandler.TestGroupStart("Equal");

            Matrix2x2 m = Matrix2x2.Create(0.1, 1 / 7, -6, -0.9);

            Assert.AssertEqual(m == Matrix2x2.Create(0.1, 1 / 7, -6, -0.9), true);
            Assert.AssertEqual(m != Matrix2x2.Create(0.1, 1 / 7, -6, -0.9), false);
            Assert.AssertEqual(Matrix2x2.CreateIdentity().Equals(Matrix2x2.CreateIdentity()), true);

            testHandler.TestGroupEnd();
        }
    }
}
