﻿using SimpleVectorMatrixPlayground.VectorMatrix;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SimpleVectorMatrixPlayground.Tests
{
    class Vector2Tests
    {
        private readonly ITestHandler testHandler;
        private readonly AssertionUtil Assert;

        public Vector2Tests(ITestHandler ts) {
            testHandler = ts;
            Assert = new AssertionUtil(ts);
        }

        public void Run()
        {
            testHandler.TestSuiteStart("Vector2Tests");

            RunLengthTests();
            RunBasicArithmeticTests();
            RunNormalizeTests();
            RunDotProductTests();
            RunPerpDotProductTests();
            RunNegatePerpendicularTests();
            RunInterpolationTests();
            RunAbsOrthogonalColinearTests();
            RunRotationTests();
            RunShearingScalingTests();
            RunProjectionTests();
            RunReflectionTests();
            RunPolarTests();

            testHandler.TestSuiteEnd();
        }

        private void RunLengthTests()
        {
            testHandler.TestGroupStart("Length / LengthSquared / Distance");

            Assert.AssertEqual(Vector2.Create().GetLength(), 0);
            Assert.AssertEqual(Vector2.Create(0, 1).GetLength(), 1);
            Assert.AssertEqual(Vector2.Create(1, 0).GetLength(), 1);
            Assert.AssertEqual(Vector2.Create(3, 4).GetLength(), 5);

            Assert.AssertEqual(Vector2.Create().GetLengthSquared(), 0);
            Assert.AssertEqual(Vector2.Create(1, 1).GetLengthSquared(), 2);
            Assert.AssertEqual(Vector2.Create(3, 4).GetLengthSquared(), 25);

            Assert.AssertEqual(Vector2.GetDistance(Vector2.Create(), Vector2.Create()), 0);
            Assert.AssertEqual(Vector2.GetDistance(Vector2.Create(1, 0), Vector2.Create()), 1);
            Assert.AssertEqual(Vector2.GetDistance(Vector2.Create(), Vector2.Create(1, 0)), 1);
            Assert.AssertEqual(Vector2.GetDistance(Vector2.Create(-1, -2), Vector2.Create(2, 2)), 5);

            testHandler.TestGroupEnd();
        }

        private void RunBasicArithmeticTests()
        {
            testHandler.TestGroupStart("Basic Arithmetic");

            Assert.AssertEqual(Vector2.Create() + Vector2.Create(), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(1, 0) + Vector2.Create(0, 1), Vector2.Create(1, 1));
            Assert.AssertEqual(Vector2.Create(-1, 0) + Vector2.Create(-2, 8), Vector2.Create(-3, 8));
            Assert.AssertEqual(Vector2.Create(-10.5, 344) + Vector2.Create(10.5, 344), Vector2.Create(0, 688));

            Assert.AssertEqual(Vector2.Create() - Vector2.Create(), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(1, 1) - Vector2.Create(1, 1), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(1, 0) - Vector2.Create(0, 1), Vector2.Create(1, -1));
            Assert.AssertEqual(Vector2.Create(8, -4) - Vector2.Create(-2, -1), Vector2.Create(10, -3));
            Assert.AssertEqual(Vector2.Create(-11, 4) - Vector2.Create(-2, -1), Vector2.Create(-9, 5));

            Assert.AssertEqual(Vector2.Create(1, -5) * 0, Vector2.Create());
            Assert.AssertEqual(Vector2.Create(1, -5) * 2, Vector2.Create(2, -10));
            Assert.AssertEqual(Vector2.Create(3, -3) * 1.5, Vector2.Create(4.5, -4.5));

            Assert.AssertEqual(Vector2.Create() / 1, Vector2.Create());
            Assert.AssertEqual(Vector2.Create(10, 12) / 2, Vector2.Create(5, 6));
            Assert.AssertEqual(Vector2.Create(25, -17.5) / 2.5, Vector2.Create(10, -7));

            testHandler.TestGroupEnd();
        }

        private void RunNormalizeTests()
        {
            testHandler.TestGroupStart("Normalize");

            Assert.AssertEqual(Vector2.Create(1, 0).GetNormalized(), Vector2.Create(1, 0));
            Assert.AssertEqual(Vector2.Create(1, 0).GetNormalized().GetLength(), 1);
            Assert.AssertEqual(Vector2.Create(0, 1).GetNormalized(), Vector2.Create(0, 1));
            Assert.AssertEqual(Vector2.Create(0, 1).GetNormalized().GetLength(), 1);

            Assert.AssertEqual(Vector2.Create(3, 4).GetNormalized(), Vector2.Create(3 / 5.0, 4 / 5.0));
            Assert.AssertEqual(Vector2.Create(3, 4).GetNormalized().GetLength(), 1);
            Assert.AssertEqual(Vector2.Create(-3, 4).GetNormalized(), Vector2.Create(-3 / 5.0, 4 / 5.0));
            Assert.AssertEqual(Vector2.Create(-3, 4).GetNormalized().GetLength(), 1);

            Assert.AssertEqual(Vector2.Create(12, -5).GetNormalized(), Vector2.Create(12 / 13.0, -5 / 13.0));
            Assert.AssertEqual(Vector2.Create(12, -5).GetNormalized().GetLength(), 1);
            Assert.AssertEqual(Vector2.Create(-5, -12).GetNormalized(), Vector2.Create(-5 / 13.0, -12 / 13.0));
            Assert.AssertEqual(Vector2.Create(-5, -12).GetNormalized().GetLength(), 1);

            Assert.AssertEqual(Double.IsNaN(Vector2.Create().GetNormalized().X), true);
            Assert.AssertEqual(Double.IsNaN(Vector2.Create().GetNormalized().Y), true);
            Assert.AssertEqual(Double.IsNaN(Vector2.Create().GetNormalized().GetLength()), true);

            Assert.AssertEqual(Vector2.Create(1, 1).GetNormalized().GetLength(), 1,  0.001);

            testHandler.TestGroupEnd();
        }

        private void RunDotProductTests()
        {
            testHandler.TestGroupStart("DotProduct");

            // perpendicular
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(), Vector2.Create()), 0);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(1, 0), Vector2.Create(0, 1)), 0);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(-2, 3), Vector2.Create(3, 2)), 0);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(-2, 3), Vector2.Create(-3, -2)), 0);

            // parallel
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(1, 0), Vector2.Create(-1, 0)), -1);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(3, 4), Vector2.Create(3, 4)), 9 + 16);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(-3, -4), Vector2.Create(3, 4)), -9 + -16);

            // arbitrary
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(1, 1), Vector2.Create(1, 0)), Math.Sqrt(2) * 1 * Math.Cos(Math.PI / 4), 0.0001);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(1, 1), Vector2.Create(0, 1)), Math.Sqrt(2) * 1 * Math.Cos(Math.PI / 4), 0.0001);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(-1, -1), Vector2.Create(1, 0)), -Math.Sqrt(2) * 1 * Math.Cos(Math.PI / 4), 0.0001);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(-1, -1), Vector2.Create(0, 1)), -Math.Sqrt(2) * 1 * Math.Cos(Math.PI / 4), 0.0001);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(-1, -1), Vector2.Create(-1, 0)), Math.Sqrt(2) * 1 * Math.Cos(Math.PI / 4), 0.0001);
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(-1, -1), Vector2.Create(0, -1)), Math.Sqrt(2) * 1 * Math.Cos(Math.PI / 4), 0.0001);

            testHandler.TestGroupEnd();
        }

        private void RunPerpDotProductTests()
        {
            testHandler.TestGroupStart("PerpDotProduct");

            // parallel
            Assert.AssertEqual(Vector2.DotProduct(Vector2.Create(), Vector2.Create()), 0);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(5, 0), Vector2.Create(4, 0)), 0);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(1, 0), Vector2.Create(-1, 0)), 0);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(3, 4), Vector2.Create(3, 4)), 0);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(-3, -4), Vector2.Create(3, 4)), 0);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(1, 2), Vector2.Create(-8, -16)), 0);

            // perpendicular
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(1, 0), Vector2.Create(0, 1)), 1);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(2, 0), Vector2.Create(0, 3)), 6);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(-2, 0), Vector2.Create(0, 3)), -6);

            // > 0 if b is counterclockwise from a -> "left turn"
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(2, 1), Vector2.Create(1, 3)), 5);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(-2, 1), Vector2.Create(-1, 0)), 1);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(-8, 1), Vector2.Create(-16, 1)), 8);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(1, 0), Vector2.Create(1, 1)), 1);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(20, 30), Vector2.Create(45, 70)), 50);

            // < 0 if b is clockwise from a -> "right turn"
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(2, 2), Vector2.Create(3, 1)), -4);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(1, 0), Vector2.Create(1, -0.5)), -0.5);
            Assert.AssertEqual(Vector2.PerpDotProduct(Vector2.Create(-4, -1), Vector2.Create(-2, 1)), -6);

            testHandler.TestGroupEnd();
        }

        private void RunNegatePerpendicularTests()
        {
            testHandler.TestGroupStart("Negate / Perpendicular vectors");

            // negate
            Assert.AssertEqual(Vector2.Create().Negate(), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(1, 1).Negate(), Vector2.Create(-1, -1));
            Assert.AssertEqual(Vector2.Create(1, 0).Negate(), Vector2.Create(-1, 0));
            Assert.AssertEqual(Vector2.Create(-6, -8).Negate(), Vector2.Create(6, 8));
            Assert.AssertEqual(Vector2.Create(1, -3).Negate(), Vector2.Create(-1, 3));

            // get perpendicular vector counterclockwise
            Assert.AssertEqual(Vector2.Create().PerpendicularCounterClockwise(), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(1, 2).PerpendicularCounterClockwise(), Vector2.Create(-2, 1));
            Assert.AssertEqual(Vector2.Create(1, 0).PerpendicularCounterClockwise(), Vector2.Create(0, 1));
            Assert.AssertEqual(Vector2.Create(0, 1).PerpendicularCounterClockwise(), Vector2.Create(-1, 0));
            Assert.AssertEqual(Vector2.Create(0, -2).PerpendicularCounterClockwise(), Vector2.Create(2, 0));
            Assert.AssertEqual(Vector2.Create(-3.5, 0).PerpendicularCounterClockwise(), Vector2.Create(0, -3.5));

            // get perpendicular vector clockwise
            Assert.AssertEqual(Vector2.Create().PerpendicularClockwise(), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(1, 2).PerpendicularClockwise(), Vector2.Create(2, -1));
            Assert.AssertEqual(Vector2.Create(1, 0).PerpendicularClockwise(), Vector2.Create(0, -1));
            Assert.AssertEqual(Vector2.Create(0, 1).PerpendicularClockwise(), Vector2.Create(1, 0));
            Assert.AssertEqual(Vector2.Create(-1, 0).PerpendicularClockwise(), Vector2.Create(0, 1));
            Assert.AssertEqual(Vector2.Create(0, -1).PerpendicularClockwise(), Vector2.Create(-1, 0));

            testHandler.TestGroupEnd();
        }

        private void RunInterpolationTests()
        {
            testHandler.TestGroupStart("Interpolation");

            Assert.AssertEqual(Vector2.Create().GetInterpolation(Vector2.Create(), 0), Vector2.Create());
            Assert.AssertEqual(Vector2.Create().GetInterpolation(Vector2.Create(), -1), Vector2.Create());
            Assert.AssertEqual(Vector2.Create().GetInterpolation(Vector2.Create(), 90), Vector2.Create());

            Assert.AssertEqual(Vector2.Create().GetInterpolation(Vector2.Create(1, 1), 0.5), Vector2.Create(0.5, 0.5));
            Assert.AssertEqual(Vector2.Create(1, 0).GetInterpolation(Vector2.Create(-3, 0), 0.5), Vector2.Create(-1, 0));
            Assert.AssertEqual(Vector2.Create(0, -4).GetInterpolation(Vector2.Create(0, 2), 0.25), Vector2.Create(0, -2.5));

            Assert.AssertEqual(Vector2.Create(212, -3232).GetInterpolation(Vector2.Create(1221,24), 0), Vector2.Create(212, -3232));
            Assert.AssertEqual(Vector2.Create(-66212, -1.12).GetInterpolation(Vector2.Create(433453, -2334.1), 1), Vector2.Create(433453, -2334.1));

            Assert.AssertEqual(Vector2.Create(1,1).GetInterpolation(Vector2.Create(2,2), -1), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(-5, -5).GetInterpolation(Vector2.Create(), 3), Vector2.Create(10, 10));

            testHandler.TestGroupEnd();
        }

        private void RunAbsOrthogonalColinearTests()
        {
            testHandler.TestGroupStart("Absolute / Orthogonal / Colinear");

            Assert.AssertEqual(Vector2.Create().GetAbsolute(), Vector2.Create());
            Assert.AssertEqual(Vector2.Create(-1.2, 6).GetAbsolute(), Vector2.Create(1.2, 6));
            Assert.AssertEqual(Vector2.Create(-2.6, -4).GetAbsolute(), Vector2.Create(2.6, 4));
            Assert.AssertEqual(Vector2.Create(0.4, -3).GetAbsolute(), Vector2.Create(0.4, 3));
            Assert.AssertEqual(Vector2.Create(7, 8.7).GetAbsolute(), Vector2.Create(7, 8.7));

            Assert.AssertEqual(Vector2.Create(1, 0).IsOrthogonal(Vector2.Create(0, 1)), true);
            Assert.AssertEqual(Vector2.Create(1, 0).IsOrthogonal(Vector2.Create(0, -1)), true);
            Assert.AssertEqual(Vector2.Create(1, 0).IsOrthogonal(Vector2.Create(-1, 0)), false);
            Assert.AssertEqual(Vector2.Create(1, 0).IsOrthogonal(Vector2.Create(1, 1)), false);

            Assert.AssertEqual(Vector2.Create(1, 0).IsColinear(Vector2.Create(2, 0)), true);
            Assert.AssertEqual(Vector2.Create(1, 0).IsColinear(Vector2.Create(-2, 0)), true);
            Assert.AssertEqual(Vector2.Create(-2, -2).IsColinear(Vector2.Create(5, 5)), true);
            Assert.AssertEqual(Vector2.Create(1, 0).IsColinear(Vector2.Create(1, 1)), false);

            testHandler.TestGroupEnd();
        }

        private void RunRotationTests()
        {
            testHandler.TestGroupStart("Rotation");

            Assert.AssertEqual(Vector2.Create(1, 0).Rotate(0), Vector2.Create(1, 0));
            Assert.AssertEqual(Vector2.Create(1, 0).Rotate(Math.PI / 2), Vector2.Create(0, 1), 0.0001);
            Assert.AssertEqual(Vector2.Create(1, 0).Rotate(Math.PI), Vector2.Create(-1, 0), 0.0001);
            Assert.AssertEqual(Vector2.Create(1, 0).Rotate(Math.PI * 1.5), Vector2.Create(0, -1), 0.0001);
            Assert.AssertEqual(Vector2.Create(1, 0).Rotate(Math.PI * 2), Vector2.Create(1, 0), 0.0001);

            Assert.AssertEqual(Vector2.Create(1, 0).Rotate(-Math.PI / 2), Vector2.Create(0, -1), 0.0001);
            Assert.AssertEqual(Vector2.Create(1, 0).Rotate(-Math.PI), Vector2.Create(-1, 0), 0.0001);

            Assert.AssertEqual(Vector2.Create(1, 1).Rotate(Math.PI / 4), Vector2.Create(0, Math.Sqrt(2)), 0.0001);

            testHandler.TestGroupEnd();
        }

        private void RunShearingScalingTests()
        {
            testHandler.TestGroupStart("Shearing / Scaling");

            Assert.AssertEqual(Vector2.Create(1, 1).Shear(2, 1), Vector2.Create(3, 2));
            Assert.AssertEqual(Vector2.Create(2, 2).Shear(2, 1), Vector2.Create(6, 4));
            Assert.AssertEqual(Vector2.Create(1, 0).Shear(0, 8), Vector2.Create(0, 1));
            Assert.AssertEqual(Vector2.Create(1, 0).Shear(8, 0), Vector2.Create(8, 1));

            Assert.AssertEqual(Vector2.Create(1, 0).Scale(8, 2), Vector2.Create(8, 0));
            Assert.AssertEqual(Vector2.Create(-2, -3).Scale(4, -1.5), Vector2.Create(-8, 4.5));
            Assert.AssertEqual(Vector2.Create(4, 2).Scale(0.5, 0.25), Vector2.Create(2, 0.5));

            testHandler.TestGroupEnd();
        }

        private void RunProjectionTests()
        {
            testHandler.TestGroupStart("Projection");

            Assert.AssertEqual(Vector2.GetProjectedVectorLength(Vector2.Create(5, 0), Vector2.Create(3, 2)), 3);
            Assert.AssertEqual(Vector2.GetProjectedVectorLength(Vector2.Create(0, 1), Vector2.Create(4, 4)), 4);
            Assert.AssertEqual(Vector2.GetProjectedVectorLength(Vector2.Create(1, 0), Vector2.Create(4, 4)), 4);
            Assert.AssertEqual(Vector2.GetProjectedVectorLength(Vector2.Create(0, 1), Vector2.Create(-4, 0)), 0);
            Assert.AssertEqual(Vector2.GetProjectedVectorLength(Vector2.Create(1, 1), Vector2.Create(-1, 1)), 0);
            Assert.AssertEqual(Vector2.GetProjectedVectorLength(Vector2.Create(1, 1), Vector2.Create(-1, -1)), -Vector2.Create(-1, -1).GetLength(), 0.0001);
            Assert.AssertEqual(Vector2.GetProjectedVectorLength(Vector2.Create(-1, -1), Vector2.Create(1, 1)), -Vector2.Create(-1, -1).GetLength(), 0.0001);

            Assert.AssertEqual(Vector2.GetProjectedVector(Vector2.Create(5, 0), Vector2.Create(3, 2)), Vector2.Create(3, 0));
            Assert.AssertEqual(Vector2.GetProjectedVector(Vector2.Create(0, 1), Vector2.Create(4, 4)), Vector2.Create(0, 4));
            Assert.AssertEqual(Vector2.GetProjectedVector(Vector2.Create(0, 1), Vector2.Create(-4, 0)), Vector2.Create());
            Assert.AssertEqual(Vector2.GetProjectedVector(Vector2.Create(1, 1), Vector2.Create(-1, -1)), Vector2.Create(-1, -1), 0.0001);
            Assert.AssertEqual(Vector2.GetProjectedVector(Vector2.Create(-1, -1), Vector2.Create(1, 1)), Vector2.Create(1, 1), 0.0001);
            Assert.AssertEqual(Vector2.GetProjectedVector(Vector2.Create(-2, -1), Vector2.Create(1, 1)), Vector2.Create(1.2, 0.6), 0.0001);
            Assert.AssertEqual(Vector2.GetProjectedVector(Vector2.Create(1, 1), Vector2.Create(-2, -1)), Vector2.Create(-1.5, -1.5), 0.0001);

            testHandler.TestGroupEnd();
        }

        private void RunReflectionTests()
        {
            testHandler.TestGroupStart("Reflection");

            Assert.AssertEqual(Vector2.GetReflectedVector(Vector2.Create(1, -1), Vector2.Create(0, 1)), Vector2.Create(1, 1));
            Assert.AssertEqual(Vector2.GetReflectedVector(Vector2.Create(-1, -1), Vector2.Create(0, 1)), Vector2.Create(-1, 1));
            Assert.AssertEqual(Vector2.GetReflectedVector(Vector2.Create(-1, -1), Vector2.Create(1, 0)), Vector2.Create(1, -1));

            Assert.AssertEqual(Vector2.GetReflectedVector(Vector2.Create(5, -5), Vector2.Create(0, 1)), Vector2.Create(5, 5));
            Assert.AssertEqual(Vector2.GetReflectedVector(Vector2.Create(4, 1), Vector2.Create(1, 2).Normalize()), Vector2.Create(1.6, -3.8));

            testHandler.TestGroupEnd();
        }

        private void RunPolarTests()
        {
            testHandler.TestGroupStart("Polar");

            Assert.AssertEqual(Vector2.GetVectorFromPolar(0, 0), Vector2.Create());
            Assert.AssertEqual(Vector2.GetVectorFromPolar(5, 0), Vector2.Create(5, 0), 0.00001);
            Assert.AssertEqual(Vector2.GetVectorFromPolar(5, Math.PI / 2), Vector2.Create(0, 5), 0.00001);
            Assert.AssertEqual(Vector2.GetVectorFromPolar(5, Math.PI), Vector2.Create(-5, 0), 0.00001);
            Assert.AssertEqual(Vector2.GetVectorFromPolar(5, Math.PI * 1.5), Vector2.Create(0, -5), 0.00001);
            Assert.AssertEqual(Vector2.GetVectorFromPolar(Math.Sqrt(32), Math.PI / 4), Vector2.Create(4, 4), 0.0001);

            Assert.AssertEqual(Vector2.Create(1,1).GetPolarAngle(), Math.PI / 4, 0.00001);
            Assert.AssertEqual(Vector2.Create(-1, 0).GetPolarAngle(), - Math.PI / 2, 0.00001);
            Assert.AssertEqual(Vector2.Create(1, 0).GetPolarAngle(), Math.PI / 2, 0.00001);
            Assert.AssertEqual(Vector2.Create(0, -1).GetPolarAngle(), Math.PI,  0.00001);
            Assert.AssertEqual(Vector2.Create(-1, -1).GetPolarAngle(), -Math.PI / 4 * 3, 0.00001);

            testHandler.TestGroupEnd();
        }



    }

}
