﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SimpleVectorMatrixPlayground.VectorMatrix
{
    public class Matrix2x2
    {
        /// <summary>
        /// Row-major stored elements 2x2 matrix.
        /// Chosen to use concrete four single elements m11, m12, m21 and m22 instead of a two dimensional array
        /// as the matrix is quite small.
        ///     | m11 m12 |
        /// M = | m21 m22 |
        /// </summary>
        private double m11, m12;
        private double m21, m22;

        private Matrix2x2() { }

        /// <summary>
        /// Create a new zero matrix.
        /// </summary>
        /// <returns></returns>
        public static Matrix2x2 Create()
        {
            return new Matrix2x2();
        }

        /// <summary>
        /// Creates a new 2x2 matrix from a two dimensional (row-major) matrix.
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static Matrix2x2 Create(double[,] matrix)
        {
            Matrix2x2 m = new Matrix2x2();
            m.m11 = matrix[0, 0];
            m.m12 = matrix[0, 1];
            m.m21 = matrix[1, 0];
            m.m22 = matrix[1, 1];
            return m;
        }

        /// <summary>
        /// Creates a new 2x2 matrix from four separate elements.
        /// </summary>
        /// <param name="m11"></param>
        /// <param name="m12"></param>
        /// <param name="m21"></param>
        /// <param name="m22"></param>
        /// <returns></returns>
        public static Matrix2x2 Create(double m11, double m12, double m21, double m22)
        {
            Matrix2x2 m = new Matrix2x2();
            m.m11 = m11;
            m.m12 = m12;
            m.m21 = m21;
            m.m22 = m22;
            return m;
        }

        /// <summary>
        /// Creates a new 2x2 identity matrix.
        /// </summary>
        /// <returns></returns>
        public static Matrix2x2 CreateIdentity()
        {
            Matrix2x2 m = new Matrix2x2();
            m.m11 = 1;
            m.m22 = 1;
            return m;
        }

        /// <summary>
        /// Zero-based(!) index access to the single elements of the matrix.
        /// 0: m11 (first row, first column)
        /// 1: m12 (first row, second column)
        /// 2: m21 (second row, first column)
        /// 3: m22 (second row, second column)
        /// </summary>
        /// <param name="elem">element to access in range [0,3]</param>
        /// <returns></returns>
        public double this[int elem]
        {
            get
            {
                switch (elem)
                {
                    case 0: return m11;
                    case 1: return m12;
                    case 2: return m21;
                    case 3: return m22;
                    default: throw new ArgumentOutOfRangeException("Invalid access index");
                }
            }

            set
            {
                switch (elem)
                {
                    case 0: m11 = value; break;
                    case 1: m12 = value; break;
                    case 2: m21 = value; break;
                    case 3: m22 = value; break;
                    default: throw new ArgumentOutOfRangeException("Invalid access index");
                }
            }
        }

        /// <summary>
        /// Zero-based(!) index access to the single elements of the matrix
        /// </summary>
        /// <param name="row"></param>
        /// <param name="col"></param>
        /// <returns></returns>
        public double this[int row, int col]
        {
            get {
                if (row < 0 || col < 0 || row >= 2 || col >= 2)
                    throw new ArgumentOutOfRangeException("Invalid access index");
                return this[row * 2 + col];
            }
            internal set {
                if (row < 0 || col < 0 || row >= 2 || col >= 2)
                    throw new ArgumentOutOfRangeException("Invalid access index");
                this[row * 2 + col] = value;
            }
        }

        public double[] GetRow(int row)
        {
            switch (row)
            {
                case 0: return new double[2] { m11, m12 };
                case 1: return new double[2] { m21, m22 };
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public Vector2 GetRowVector(int row)
        {
            switch (row)
            {
                case 0: return Vector2.Create(m11, m12);
                case 1: return Vector2.Create(m21, m22);
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public double[] GetColumn(int col)
        {
            switch (col)
            {
                case 0: return new double[2] { m11, m21 };
                case 1: return new double[2] { m12, m22 };
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public Vector2 GetColumnVector(int col)
        {
            switch (col)
            {
                case 0: return Vector2.Create(m11, m21);
                case 1: return Vector2.Create(m12, m22);
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            for(var i = 0; i < 2; i++)
            {
                for (var j = 0; j < 2; j++)
                {
                    sb.Append("| ");
                    sb.Append(this[i, j]).Append(' ');
                }
            
                sb.Append("|").AppendLine();
            }
            
            return sb.ToString();
        }

        public static Matrix2x2 operator +(Matrix2x2 a, Matrix2x2 b)
        {
            return Matrix2x2.Add(a, b);
        }

        public static Matrix2x2 Add(Matrix2x2 a, Matrix2x2 b)
        {
            Matrix2x2 c = Matrix2x2.Create();
            c.m11 = a.m11 + b.m11;
            c.m12 = a.m12 + b.m12;
            c.m21 = a.m21 + b.m21;
            c.m22 = a.m22 + b.m22;

            /* alternatively more generic */
            //for (int row = 0; row < 2; row++)
            //{
            //    for (int col = 0; col < 2; col++)
            //    {
            //        c[row, col] = a[row, col] + b[row, col];
            //    }
            //}
            return c;
        }

        public static Matrix2x2 operator -(Matrix2x2 a, Matrix2x2 b)
        {
            return Matrix2x2.Sub(a, b);
        }

        public static Matrix2x2 Sub(Matrix2x2 a, Matrix2x2 b)
        {
            Matrix2x2 c = Matrix2x2.Create();
            c.m11 = a.m11 - b.m11;
            c.m12 = a.m12 - b.m12;
            c.m21 = a.m21 - b.m21;
            c.m22 = a.m22 - b.m22;

            /* alternatively more generic */
            //for (int row = 0; row < 2; row++)
            //{
            //    for (int col = 0; col < 2; col++)
            //    {
            //        c[row, col] = a[row, col] - b[row, col];
            //    }
            //}

            return c;
        }

        public static Matrix2x2 operator *(Matrix2x2 a, double b)
        {
            return Matrix2x2.Mul(a, b);
        }

        public static Matrix2x2 Mul(Matrix2x2 a, double b)
        {
            Matrix2x2 c = Matrix2x2.Create();
            c.m11 = a.m11 * b;
            c.m12 = a.m12 * b;
            c.m21 = a.m21 * b;
            c.m22 = a.m22 * b;

            /* alternatively more generic */
            //for (int row = 0; row < 2; row++)
            //{
            //    for (int col = 0; col < 2; col++)
            //    {
            //        c[row, col] = a[row, col] * b;
            //    }
            //}

            return c;
        }

        public static Matrix2x2 operator *(Matrix2x2 a, Matrix2x2 b)
        {
            double m11 = a.m11 * b.m11 + a.m12 * b.m21;
            double m12 = a.m11 * b.m12 + a.m12 * b.m22;

            double m21 = a.m21 * b.m11 + a.m22 * b.m21;
            double m22 = a.m21 * b.m12 + a.m22 * b.m22;

            return Matrix2x2.Create(m11, m12, m21, m22);
        }

        public Matrix2x2 Transpose()
        {
            return Matrix2x2.Create(m11, m21, m12, m22);
        }

        public Matrix2x2 Negate()
        {
            return this * -1;
        }

        /// <summary>
        /// Calculates the determinant:
        /// m11 * m22 - m12 * m21
        /// Note that this is the same as the perp dot product of either the two row vectors or the two column vectors!
        /// Thus geometrically the determinant of a 2x2 matrix calculates the (signed) area of the parallelogram of the two basis vectors.
        /// </summary>
        /// <returns></returns>
        public double Determinant()
        {
            return m11 * m22 - m12 * m21;
        }

        /// <summary>
        /// Inverse of a 2x2 matrix.
        /// Calculated as:
        ///           |  m22 -m12 |
        /// 1 / det * | -m21  m11 |
        /// </summary>
        /// <returns></returns>
        public Matrix2x2 Inverse()
        {
            double det = Determinant();
            if (det == 0)
                throw new InvalidOperationException("Determinant is zero");

            Matrix2x2 m = Matrix2x2.Create();
            double invDet = 1.0 / det;
            m[0, 0] = invDet * m22;
            m[0, 1] = -invDet * m12;
            m[1, 0] = -invDet * m21;
            m[1, 1] = invDet * m11;

            return m;
        }

        /// <summary>
        /// Create the rotation matrix with given angle.
        /// A vector that is multiplied from the right side (as column vector) is rotated counterclockwise with the given angle.
        ///     | cos(a) -sin(a) |
        /// m = | sin(a)  cos(a) |
        /// => m * v:
        ///    x' = v.x * cos(a) - v.y * sin(a)
        ///    y' = v.x * sin(a) + v.y * cos(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public static Matrix2x2 GetRotationMatrix(double angle)
        {
            Matrix2x2 mat = Matrix2x2.Create();
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            mat.m11 = cos;
            mat.m12 = -sin;
            mat.m21 = sin;
            mat.m22 = cos;
            return mat;
        }

        /// <summary>
        /// Creates the rotation matrix with given angle.
        /// A vector that is multiplied from the right side (as column vector) is rotated clockwise with the given angle.
        /// This can be derivated by rotation matrix of <see cref="GetRotationMatrix"/> and using the identities:
        /// cos(-a) = cos(a) and sin(-a) = -sin(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public static Matrix2x2 GetRotationMatrixCW(double angle)
        {
            Matrix2x2 mat = Matrix2x2.Create();
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            mat.m11 = cos;
            mat.m12 = sin;
            mat.m21 = -sin;
            mat.m22 = cos;
            return mat;
        }

        /// <summary>
        /// Multiplies a 2D vector ("column vector") from right side to the matrix.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        public static Vector2 operator *(Matrix2x2 a, Vector2 v)
        {
            Vector2 vec = Vector2.Create(a.m11 * v.X + a.m12 * v.Y, a.m21 * v.X + a.m22 * v.Y);
            return vec;
        }

        /// <summary>
        /// Multiplies a 2D vector ("row vector") from left side to the matrix.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        public static Vector2 operator *(Vector2 v, Matrix2x2 a)
        {
            Vector2 vec = Vector2.Create(a.m11 * v.X + a.m21 * v.Y, a.m12 * v.X + a.m22 * v.Y);
            return vec;
        }

        /// <summary>
        /// Create the matrix to scale a vector in x-direction and y-direction.
        /// The vector needs to be multiplied from the right side (2x1 column vector).
        ///     | sx 0 |
        /// m = | 0 sy |
        /// => m * v:
        ///    x' = v.x * sx + v.y  * 0 = v.x * sx
        ///    y' = v.x * 0 + v.y * sy = v.y * sy
        /// </summary>
        /// <param name="sx">scale factor in x-direction</param>
        /// <param name="sy">scale factor in y-direction</param>
        /// <returns></returns>
        public static Matrix2x2 GetScalingMatrix(double sx, double sy)
        {
            Matrix2x2 mat = Matrix2x2.Create();
            mat.m11 = sx;
            mat.m22 = sy;
            return mat;
        }

        /// <summary>
        /// Create the matrix to shear a vector in x-direction.
        /// The vector needs to be multiplied from the right side (2x1 column vector).
        ///     | 1 sx |
        /// m = | 0 1 |
        /// => m * v:
        ///    x' = v.x * 1 + v.y * sx = v.x + sx * v.y
        ///    y' = v.x * 0 + 1 * v.y = v.y
        /// </summary>
        /// <param name="s">shear factor</param>
        /// <returns></returns>
        public static Matrix2x2 GetShearingMatrixX(double s)
        {
            Matrix2x2 mat = Matrix2x2.Create();
            mat.m11 = 1;
            mat.m22 = 1;
            mat.m12 = s;
            return mat;
        }

        /// <summary>
        /// Create the matrix to shear a vector in y-direction.
        /// The vector needs to be multiplied from the right side (2x1 column vector).
        ///     | 1 0 |
        /// m = | sy 1 |
        /// => m * v:
        ///    x' = v.x * 1 + v.y * 0 = v.x
        ///    y' = v.x * sy + 1 * v.y = v.y + sy * v.x
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public static Matrix2x2 GetShearingMatrixY(double s)
        {
            Matrix2x2 mat = Matrix2x2.Create();
            mat.m11 = 1;
            mat.m22 = 1;
            mat.m21 = s;
            return mat;
        }

        public static bool operator !=(Matrix2x2 a, Matrix2x2 b)
        {
            return !(a == b);
        }

        public static bool operator ==(Matrix2x2 a, Matrix2x2 b)
        {
            // this is just a dummy implementation by testing equality of two matrices if elements at each position are the same
            // direct comparison of double values is dangerous
            return ((a.m11 == b.m11) && (a.m12 == b.m12) && (a.m21 == b.m21) && (a.m22 == b.m22));
        }

        public override bool Equals(object obj)
        {
            if (obj.GetType() != this.GetType())
            {
                return false;
            }

            return this.Equals(obj as Matrix2x2);
        }

        public bool Equals(Matrix2x2 obj)
        {
            if (obj is null)
            {
                return false;
            }

            if (Object.ReferenceEquals(this, obj))
            {
                return true;
            }

            return this == obj;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }
}
