﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SimpleVectorMatrixPlayground.VectorMatrix
{
    class Matrix3x3
    {
        /// <summary>
        /// Row-major stored elements 3x3 matrix.
        /// Chosen to use concrete nine single elements instead of a two dimensional array as the matrix is quite small.
        ///     | m11 m12 m13 |
        /// M = | m21 m22 m23 |
        ///     | m31 m32 m33 |
        /// </summary>
        private double m11, m12, m13;
        private double m21, m22, m23;
        private double m31, m32, m33;

        private Matrix3x3() { }

        /// <summary>
        /// Create a new zero matrix.
        /// </summary>
        /// <returns></returns>
        public static Matrix3x3 Create()
        {
            return new Matrix3x3();
        }

        /// <summary>
        /// Creates a new 2x2 matrix from a two dimensional (row-major) matrix.
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static Matrix3x3 Create(double[,] matrix)
        {
            Matrix3x3 m = new Matrix3x3();
            m.m11 = matrix[0, 0];
            m.m12 = matrix[0, 1];
            m.m13 = matrix[0, 2];
            m.m21 = matrix[1, 0];
            m.m22 = matrix[1, 1];
            m.m23 = matrix[1, 2];
            m.m31 = matrix[2, 0];
            m.m32 = matrix[2, 1];
            m.m33 = matrix[2, 2];
            return m;
        }

        /// <summary>
        /// Creates a new 3x3 matrix from nine separate elements.
        /// </summary>
        /// <param name="m11"></param>
        /// <param name="m12"></param>
        /// <param name="m21"></param>
        /// <param name="m22"></param>
        /// <returns></returns>
        public static Matrix3x3 Create(double m11, double m12, double m13,
                                       double m21, double m22, double m23,
                                       double m31, double m32, double m33)
        {
            Matrix3x3 m = new Matrix3x3();
            m.m11 = m11;
            m.m12 = m12;
            m.m13 = m13;
            m.m21 = m21;
            m.m22 = m22;
            m.m23 = m23;
            m.m31 = m31;
            m.m32 = m32;
            m.m33 = m33;
            return m;
        }

        /// <summary>
        /// Creates a new 2x2 identity matrix.
        /// </summary>
        /// <returns></returns>
        public static Matrix3x3 CreateIdentity()
        {
            Matrix3x3 m = new Matrix3x3();
            m.m11 = 1;
            m.m22 = 1;
            m.m33 = 1;
            return m;
        }

        /// <summary>
        /// Zero-based(!) index access to the single elements of the matrix.
        /// 0: m11 (first row, first column)
        /// 1: m12 (first row, second column)
        /// 2: m13 (first row, third colum)
        /// 3: m21 (second row, first column)
        /// 4: m22 (second row, second column)
        /// 5: m23 (second row, third column)
        /// 6: m31 (third row, first column)
        /// 7: m32 (third row, second column)
        /// 8: m33 (third row, third column)
        /// </summary>
        /// <param name="elem">element to access in range [0,3]</param>
        /// <returns></returns>
        public double this[int elem]
        {
            get
            {
                switch (elem)
                {
                    case 0: return m11;
                    case 1: return m12;
                    case 2: return m13;
                    case 3: return m21;
                    case 4: return m22;
                    case 5: return m23;
                    case 6: return m31;
                    case 7: return m32;
                    case 8: return m33;
                    default: throw new ArgumentOutOfRangeException("Invalid access index");
                }
            }

            set
            {
                switch (elem)
                {
                    case 0: m11 = value; break;
                    case 1: m12 = value; break;
                    case 2: m13 = value; break;
                    case 3: m21 = value; break;
                    case 4: m22 = value; break;
                    case 5: m23 = value; break;
                    case 6: m31 = value; break;
                    case 7: m32 = value; break;
                    case 8: m33 = value; break;
                    default: throw new ArgumentOutOfRangeException("Invalid access index");
                }
            }
        }

        /// <summary>
        /// Zero-based(!) index access to the single elements of the matrix
        /// </summary>
        /// <param name="row"></param>
        /// <param name="col"></param>
        /// <returns></returns>
        public double this[int row, int col]
        {
            get
            {
                if (row < 0 || col < 0 || row >= 3 || col >= 3)
                    throw new ArgumentOutOfRangeException("Invalid access index");
                return this[row * 3 + col];
            }
            internal set
            {
                if (row < 0 || col < 0 || row >= 3 || col >= 3)
                    throw new ArgumentOutOfRangeException("Invalid access index");
                this[row * 3 + col] = value;
            }
        }

        public double[] GetRow(int row)
        {
            switch (row)
            {
                case 0: return new double[3] { m11, m12, m13 };
                case 1: return new double[3] { m21, m22, m23 };
                case 2: return new double[3] { m31, m32, m33 };
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public Vector3 GetRowVector(int row)
        {
            switch (row)
            {
                case 0: return Vector3.Create(m11, m12, m13);
                case 1: return Vector3.Create(m21, m22, m23);
                case 2: return Vector3.Create(m31, m32, m33);
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public double[] GetColumn(int col)
        {
            switch (col)
            {
                case 0: return new double[3] { m11, m21, m31 };
                case 1: return new double[3] { m12, m22, m32 };
                case 2: return new double[3] { m13, m23, m33 };
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public Vector3 GetColumnVector(int col)
        {
            switch (col)
            {
                case 0: return Vector3.Create(m11, m21, m31);
                case 1: return Vector3.Create(m12, m22, m32);
                case 2: return Vector3.Create(m13, m23, m33);
                default: throw new ArgumentOutOfRangeException("Invalid access index");
            }
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            for (var i = 0; i < 3; i++)
            {
                for (var j = 0; j < 3; j++)
                {
                    sb.Append("| ");
                    sb.Append(this[i, j]).Append(' ');
                }

                sb.Append("|").AppendLine();
            }

            return sb.ToString();
        }

        public static Matrix3x3 operator +(Matrix3x3 a, Matrix3x3 b)
        {
            return Matrix3x3.Add(a, b);
        }

        public static Matrix3x3 Add(Matrix3x3 a, Matrix3x3 b)
        {
            Matrix3x3 c = Matrix3x3.Create();
            c.m11 = a.m11 + b.m11;
            c.m12 = a.m12 + b.m12;
            c.m13 = a.m13 + b.m13;
            c.m21 = a.m21 + b.m21;
            c.m22 = a.m22 + b.m22;
            c.m23 = a.m23 + b.m23;
            c.m31 = a.m31 + b.m31;
            c.m32 = a.m32 + b.m32;
            c.m33 = a.m33 + b.m33;

            /* alternatively more generic */
            //for (int row = 0; row < 3; row++)
            //{
            //    for (int col = 0; col < 3; col++)
            //    {
            //        c[row, col] = a[row, col] + b[row, col];
            //    }
            //}
            return c;
        }

        public static Matrix3x3 operator -(Matrix3x3 a, Matrix3x3 b)
        {
            return Matrix3x3.Sub(a, b);
        }

        public static Matrix3x3 Sub(Matrix3x3 a, Matrix3x3 b)
        {
            Matrix3x3 c = Matrix3x3.Create();
            c.m11 = a.m11 - b.m11;
            c.m12 = a.m12 - b.m12;
            c.m13 = a.m13 - b.m13;
            c.m21 = a.m21 - b.m21;
            c.m22 = a.m22 - b.m22;
            c.m23 = a.m23 - b.m23;
            c.m31 = a.m31 - b.m31;
            c.m32 = a.m32 - b.m32;
            c.m33 = a.m33 - b.m33;

            /* alternatively more generic */
            //for (int row = 0; row < 3; row++)
            //{
            //    for (int col = 0; col < 3; col++)
            //    {
            //        c[row, col] = a[row, col] - b[row, col];
            //    }
            //}

            return c;
        }

        public static Matrix3x3 operator *(Matrix3x3 a, double b)
        {
            return Matrix3x3.Mul(a, b);
        }

        public static Matrix3x3 Mul(Matrix3x3 a, double b)
        {
            Matrix3x3 c = Matrix3x3.Create();
            c.m11 = a.m11 * b;
            c.m12 = a.m12 * b;
            c.m13 = a.m13 * b;
            c.m21 = a.m21 * b;
            c.m22 = a.m22 * b;
            c.m23 = a.m23 * b;
            c.m31 = a.m31 * b;
            c.m32 = a.m32 * b;
            c.m33 = a.m33 * b;

            /* alternatively more generic */
            //for (int row = 0; row < 3; row++)
            //{
            //    for (int col = 0; col < 3; col++)
            //    {
            //        c[row, col] = a[row, col] * b;
            //    }
            //}

            return c;
        }

        public static Matrix3x3 operator /(Matrix3x3 a, double b)
        {
            return Matrix3x3.Div(a, b);
        }

        public static Matrix3x3 Div(Matrix3x3 a, double b)
        {
            Matrix3x3 c = Matrix3x3.Create();
            c.m11 = a.m11 / b;
            c.m12 = a.m12 / b;
            c.m13 = a.m13 / b;
            c.m21 = a.m21 / b;
            c.m22 = a.m22 / b;
            c.m23 = a.m23 / b;
            c.m31 = a.m31 / b;
            c.m32 = a.m32 / b;
            c.m33 = a.m33 / b;

            /* alternatively more generic */
            //for (int row = 0; row < 3; row++)
            //{
            //    for (int col = 0; col < 3; col++)
            //    {
            //        c[row, col] = a[row, col] / b;
            //    }
            //}

            return c;
        }

        public static Matrix3x3 operator *(Matrix3x3 a, Matrix3x3 b)
        {
            double m11 = a.m11 * b.m11 + a.m12 * b.m21 + a.m13 * b.m31;
            double m12 = a.m11 * b.m12 + a.m12 * b.m22 + a.m13 * b.m32;
            double m13 = a.m11 * b.m13 + a.m12 * b.m23 + a.m13 * b.m33;

            double m21 = a.m21 * b.m11 + a.m22 * b.m21 + a.m23 * b.m31;
            double m22 = a.m21 * b.m12 + a.m22 * b.m22 + a.m23 * b.m32;
            double m23 = a.m21 * b.m13 + a.m22 * b.m23 + a.m23 * b.m33;

            double m31 = a.m31 * b.m11 + a.m32 * b.m21 + a.m33 * b.m31;
            double m32 = a.m31 * b.m12 + a.m32 * b.m22 + a.m33 * b.m32;
            double m33 = a.m31 * b.m13 + a.m32 * b.m23 + a.m33 * b.m33;

            return Matrix3x3.Create(m11, m12, m13, m21, m22, m23, m31, m32, m33);
        }

        public Matrix3x3 Transpose()
        {
            return Matrix3x3.Create(m11, m21, m31, m12, m22, m32, m13, m23, m33);
        }

        public Matrix3x3 Negate()
        {
            return this * -1;
        }

        /// <summary>
        /// Calculates the determinant:
        /// m11 * m22 * m33 + m12 * m23 * m31 + m13 * m21 * m32
        /// - m13 * m22 * m31 - m12 * m21 * m33 - m11 * m23 * m32
        /// Geometrically the determinant of a 3x3 matrix calculates the (signed) volume of the parallelepiped of the three basis vectors.
        /// </summary>
        /// <returns></returns>
        public double Determinant()
        {
            return m11 * m22 * m33 + m12 * m23 * m31 + m13 * m21 * m32 - m13 * m22 * m31 - m12 * m21 * m33 - m11 * m23 * m32;
        }

        /// <summary>
        /// Calculates the inverse of a matrix using the adjoint of the matrix.
        /// Minv = adj(M) / |M|
        /// </summary>
        /// <returns></returns>
        public Matrix3x3 Inverse()
        {
            double det = Determinant();
            if (det == 0)
                throw new InvalidOperationException("Determinant is zero");

            double c11 = m22 * m33 - m23 * m32;
            double c12 = m21 * m33 - m23 * m31;
            double c13 = m21 * m32 - m22 * m31;

            double c21 = m12 * m33 - m13 * m32;
            double c22 = m11 * m33 - m13 * m31;
            double c23 = m11 * m32 - m12 * m31;

            double c31 = m12 * m23 - m13 * m22;
            double c32 = m11 * m23 - m13 * m21;
            double c33 = m11 * m22 - m12 * m21;

            double idet = 1.0 / det;

            Matrix3x3 m = Matrix3x3.Create(+c11 * idet, -c12 * idet, +c13 * idet,
                                           -c21 * idet, +c22 * idet, -c23 * idet,
                                           +c31 * idet, -c32 * idet, +c33 * idet).Transpose();
            return m;
        }

        /// <summary>
        /// Multiplies a 3D vector ("column vector") from right side to the matrix.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        public static Vector3 operator *(Matrix3x3 a, Vector3 v)
        {
            Vector3 vec = Vector3.Create(a.m11 * v.X + a.m12 * v.Y + a.m13 * v.Z,
                                         a.m21 * v.X + a.m22 * v.Y + a.m23 * v.Z,
                                         a.m31 * v.X + a.m32 * v.Y + a.m33 * v.Z);
            return vec;
        }

        /// <summary>
        /// Multiplies a 3D vector ("row vector") from right side to the matrix.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        public static Vector3 operator *(Vector3 v, Matrix3x3 a)
        {
            Vector3 vec = Vector3.Create(a.m11 * v.X + a.m21 * v.Y + a.m31 * v.Z,
                                         a.m12 * v.X + a.m22 * v.Y + a.m32 * v.Z,
                                         a.m13 * v.X + a.m23 * v.Y + a.m33 * v.Z);
            return vec;
        }

        /// <summary>
        /// Create the rotation matrix with given angle.
        /// A vector that is multiplied from the right side (as column vector) is rotated counterclockwise with the given angle.
        ///     | 1 0      0       |
        /// m = | 0 cos(a) -sin(a) |
        ///     | 0 sin(a)  cos(a) |
        /// => m * v:
        ///    x' = v.x
        ///    y' = v.y * cos(a) - v.z * sin(a)
        ///    z' = v.y * sin(a) + v.z * cos(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public static Matrix3x3 GetRotationMatrixX(double angle)
        {
            Matrix3x3 mat = Matrix3x3.Create();
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            mat.m11 = 1;
            mat.m22 = cos;
            mat.m23 = -sin;
            mat.m32 = sin;
            mat.m33 = cos;
            return mat;
        }

        /// <summary>
        /// Create the rotation matrix with given angle.
        /// A vector that is multiplied from the right side (as column vector) is rotated counterclockwise with the given angle.
        ///     | cos(a)  0      sin(a) |
        /// m = | 0       1      0      |
        ///     | -sin(a) 0      cos(a) |
        /// => m * v:
        ///    x' = v.x * cos(a) + v.z * sin(a)
        ///    y' = v.y
        ///    z' = -v.x * sin(a) + v.z * cos(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public static Matrix3x3 GetRotationMatrixY(double angle)
        {
            Matrix3x3 mat = Matrix3x3.Create();
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            mat.m11 = cos;
            mat.m13 = sin;
            mat.m22 = 1;
            mat.m31 = -sin;
            mat.m33 = cos;
            return mat;
        }

        /// <summary>
        /// Create the rotation matrix with given angle.
        /// A vector that is multiplied from the right side (as column vector) is rotated counterclockwise with the given angle.
        ///     | cos(a)  -sin(a) 0 |
        /// m = | sin(a)  cos(a)  0 |
        ///     | 0        0      1 |
        /// => m * v:
        ///    x' = v.x * cos(a) + v.y * -sin(a)
        ///    y' = v.y * sin(a) + v.y * cos(a)
        ///    z' = v.z
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public static Matrix3x3 GetRotationMatrixZ(double angle)
        {
            Matrix3x3 mat = Matrix3x3.Create();
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            mat.m11 = cos;
            mat.m12 = -sin;
            mat.m21 = sin;
            mat.m22 = cos;
            mat.m33 = 1;
            return mat;
        }

        /// <summary>
        /// Create the matrix to scale a vector in x-direction, y-direction and z-direction.
        /// The vector needs to be multiplied from the right side (3x1 column vector).
        ///     | sx 0  0  |
        /// m = | 0  sy 0  |
        ///     | 0  0  sz |
        /// => m * v:
        ///    x' = v.x * sx;
        ///    y' = v.y * sy;
        ///    z' = v.z * sz;
        /// </summary>
        /// <param name="sx">scale factor in x-direction</param>
        /// <param name="sy">scale factor in y-direction</param>
        /// <param name="sz">scale factor in yz-direction</param>
        /// <returns></returns>
        public static Matrix3x3 GetScalingMatrix(double sx, double sy, double sz)
        {
            Matrix3x3 mat = Matrix3x3.Create();
            mat.m11 = sx;
            mat.m22 = sy;
            mat.m33 = sz;
            return mat;
        }

        public static bool operator !=(Matrix3x3 a, Matrix3x3 b)
        {
            return !(a == b);
        }

        public static bool operator ==(Matrix3x3 a, Matrix3x3 b)
        {
            // this is just a dummy implementation by testing equality of two matrices if elements at each position are the same
            // direct comparison of double values is dangerous
            return ((a.m11 == b.m11) && (a.m12 == b.m12) && (a.m13 == b.m13) &&
                    (a.m21 == b.m21) && (a.m22 == b.m22) && (a.m23 == b.m23) &&
                    (a.m31 == b.m31) && (a.m32 == b.m32) && (a.m33 == b.m33));
        }

        public override bool Equals(object obj)
        {
            if (obj.GetType() != this.GetType())
            {
                return false;
            }

            return this.Equals(obj as Matrix3x3);
        }

        public bool Equals(Matrix3x3 obj)
        {
            if (obj is null)
            {
                return false;
            }

            if (Object.ReferenceEquals(this, obj))
            {
                return true;
            }

            return this == obj;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }
}
