﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SimpleVectorMatrixPlayground.VectorMatrix
{
    public class Vector2
    {
        public double X { private set; get; }
        public double Y { private set; get; }

        private Vector2() { }

        private Vector2(double x, double y) { this.X = x; this.Y = y; }

        public static Vector2 Create()
        {
            return new Vector2();
        }

        public static Vector2 Create(double x, double y)
        {
            return new Vector2(x, y);
        }

        public static Vector2 Create(double x1, double y1, double x2, double y2)
        {
            return new Vector2(x2 - x1, y2 - y1);
        }

        public static Vector2 Create(Vector2 v)
        {
            return new Vector2(v.X, v.Y);
        }

        public double GetLength()
        {
            return Math.Sqrt(X * X + Y * Y);
        }

        public double GetLengthSquared()
        {
            return X * X + Y * Y;
        }

        public static double GetDistance(Vector2 a, Vector2 b)
        {
            double x = a.X - b.X;
            double y = a.Y - b.Y;
            return Math.Sqrt(x * x + y * y);
        }

        public static Vector2 operator +(Vector2 a, Vector2 b)
        {
            return Vector2.Add(a, b);
        }

        public static Vector2 Add(Vector2 a, Vector2 b)
        {
            return new Vector2(a.X + b.X, a.Y + b.Y);
        }

        public static Vector2 operator -(Vector2 a, Vector2 b)
        {
            return Vector2.Sub(a, b);
        }

        public static Vector2 Sub(Vector2 a, Vector2 b)
        {
            return new Vector2(a.X - b.X, a.Y - b.Y);
        }

        public static Vector2 operator *(Vector2 a, double b)
        {
            return Vector2.Mul(a, b);
        }

        public static Vector2 operator *(double b, Vector2 a)
        {
            return Vector2.Mul(a, b);
        }

        public static Vector2 Mul(Vector2 a, double b)
        {
            return new Vector2(a.X * b, a.Y * b);
        }

        public static Vector2 operator /(Vector2 a, double b)
        {
            return Vector2.Div(a, b);
        }

        public static Vector2 Div(Vector2 a, double b)
        {
            return new Vector2(a.X / b, a.Y / b);
        }

        /// <summary>
        /// Normalize a vector by dividing each component by the vector length / magnitude. The resulting vector has always length 1.
        /// </summary>
        /// <returns></returns>
        public Vector2 GetNormalized()
        {
            return Vector2.Create(this).Normalize();
        }

        public Vector2 Normalize()
        {
            double len = GetLength();
            this.X /= len;
            this.Y /= len;
            return this;
        }

        /// <summary>
        /// Calculates the dot product of two vectors.
        /// DotProduct(a, b) = |a| * |b| * cos(alpha).
        /// If both vectors are orthogonal, the dot product is zero.
        /// If both vectors are parallel, the dot product is the product of the length of the vectors (because cos(0) = 1).
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static double DotProduct(Vector2 a, Vector2 b)
        {
            return ((a.X * b.X) + (a.Y * b.Y));
        }

        public double DotProduct(Vector2 a)
        {
            return DotProduct(this, a);
        }

        /// <summary>
        /// Calculatres the PerpDot product of vectors a and b which is calculated as the dot product of the perpendicular vector of a and vector b.
        /// PerpDotProduct(a, b) = |a| * |b| * sin(alpha) which is the area of the parallelogram of both vectors.
        /// If both vectors are orthogonal, the PerpDot product is the area of the rectangle, thus |a| * |b|.
        /// If both vectors are parallel, the PerpDot product is zero.
        /// If the PerpDot product is > 0, then b is counterclockwise from a -> "left turn".
        /// If the PerpDot product is < 0, then b is clockwise from a -> "right turn".
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static double PerpDotProduct(Vector2 a, Vector2 b)
        {
            return ((a.X * b.Y) - (a.Y * b.X));
        }

        public double PerpDotProduct(Vector2 a)
        {
            return PerpDotProduct(this, a);
        }

        /// <summary>
        /// Returns the vector with the same length that points into the opposite direction.
        /// </summary>
        /// <returns></returns>
        public Vector2 GetNegated()
        {
            return Vector2.Create(this).Negate();
        }

        public Vector2 Negate()
        {
            this.X = -this.X;
            this.Y = -this.Y;
            return this;
        }

        /// <summary>
        /// Gets the vector that is perpendicular to the given vector in counter clockwise direction.
        /// </summary>
        /// <returns></returns>
        public Vector2 GetPerpendicularCounterClockwise()
        {
            return Vector2.Create(this).PerpendicularCounterClockwise();
        }

        public Vector2 PerpendicularCounterClockwise()
        {
            double x = this.X;
            this.X = -this.Y;
            this.Y = x;
            return this;
        }

        /// <summary>
        /// Gets the vector that is perpendicular to the given vector in clockwise direction.
        /// </summary>
        /// <returns></returns>
        public Vector2 GetPerpendicularClockwise()
        {
            return Vector2.Create(this).PerpendicularClockwise();
        }

        public Vector2 PerpendicularClockwise()
        {
            double x = this.X;
            this.X = this.Y;
            this.Y = -x;
            return this;
        }

        /// <summary>
        /// Gets the linear interpolation between two vectors.
        /// return = a + (b - a) * t.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="t"></param>
        /// <returns></returns>
        public static Vector2 GetInterpolation(Vector2 a, Vector2 b, double t)
        {
            double dx = b.X - a.X;
            double dy = b.Y - a.Y;

            return Vector2.Create(a.X + t * dx, a.Y + t * dy);
        }

        public Vector2 GetInterpolation(Vector2 a, double t)
        {
            return Vector2.GetInterpolation(this, a, t);
        }

        /// <summary>
        /// Gets the vector with absolute coordinate values.
        /// </summary>
        /// <returns></returns>
        public Vector2 GetAbsolute()
        {
            return Vector2.Create(this.X < 0 ? -this.X : this.X, this.Y < 0 ? -this.Y : this.Y);
        }

        /// <summary>
        /// Returns true if a and b are orthogonal / perpendicular, otherwise false
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static bool IsOrthogonal(Vector2 a, Vector2 b)
        {
            return DotProduct(a, b) == 0;
        }

        public bool IsOrthogonal(Vector2 a)
        {
            return DotProduct(this, a) == 0;
        }


        /// <summary>
        /// Returns true if a and b are colinear, otherwise false
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static bool IsColinear(Vector2 a, Vector2 b)
        {
            return PerpDotProduct(a, b) == 0;
        }

        public bool IsColinear(Vector2 a)
        {
            return PerpDotProduct(this, a) == 0;
        }

        /// <summary>
        /// Gets the vector rotated by given angle in counterclockwise direction.
        /// x = x * cos(a) - y * sin(a)
        /// y = y * sin(a) + y * cos(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public Vector2 GetRotated(double angle)
        {
            return Vector2.Create(this).Rotate(angle);
        }

        public Vector2 Rotate(double angle)
        {
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            double newX = this.X * cos - this.Y * sin;
            double newY = this.X * sin + this.Y * cos;
            this.X = newX;
            this.Y = newY;
            return this;
        }

        /// <summary>
        /// Get the sheared vector.
        /// x = x * xs + y
        /// y = y * ys + x
        /// </summary>
        /// <param name="xs"></param>
        /// <param name="ys"></param>
        /// <returns></returns>
        public Vector2 GetSheared(double xs, double ys)
        {
            return Vector2.Create(this).Shear(xs, ys);
        }

        public Vector2 Shear(double xs, double ys)
        {
            double newX = this.X * xs + this.Y;
            double newY = this.Y * ys + this.X;
            this.X = newX;
            this.Y = newY;
            return this;
        }

        /// <summary>
        /// Gets the non-uniform scaled vector.
        /// </summary>
        /// <param name="xs"></param>
        /// <param name="ys"></param>
        /// <returns></returns>
        public Vector2 GetScaled(double xs, double ys)
        {
            return Vector2.Create(this).Scale(xs, ys);
        }

        public Vector2 Scale(double xs, double ys)
        {
            this.X = this.X * xs;
            this.Y = this.Y * ys;
            return this;
        }

        /// <summary>
        /// Get the length of the vector when Vector b is projected onto vector a.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static double GetProjectedVectorLength(Vector2 a, Vector2 b)
        {
            return DotProduct(a, b) / a.GetLength();
        }

        /// <summary>
        /// Get the length of the vector when Vector v is projected onto current vector.
        /// </summary>
        /// <param name="a"></param>
        /// <returns></returns>
        public double GetProjectedVectorLength(Vector2 v)
        {
            return DotProduct(this, v) / v.GetLength();
        }

        /// <summary>
        /// Get the result vector when Vector b is projected onto vector a.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static Vector2 GetProjectedVector(Vector2 a, Vector2 b)
        {
            return a.GetNormalized() *  DotProduct(a, b) / a.GetLength();
        }

        /// <summary>
        /// Get the result vector when Vector v is projected onto current vector.
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public Vector2 GetProjectedVector(Vector2 v)
        {
            return this.GetNormalized() * DotProduct(this, v) / this.GetLength();
        }

        /// <summary>
        /// Get the reflected vector a at the surface given with normal n.
        /// </summary>
        /// <param name="a">The vector a to reflect at the surface</param>
        /// <param name="n">Normal of the surface. Bit be normalited (length = 1) of course-</param>
        /// <returns></returns>
        public static Vector2 GetReflectedVector(Vector2 a, Vector2 n)
        {
            Vector2 r = a - (2 * DotProduct(a, n) * n);
            return r;
        }

        public static Vector2 GetVectorFromPolar(double mag, double angle)
        {
            return Vector2.Create(mag* Math.Cos(angle), mag * Math.Sin(angle));
        }

        public double GetPolarMagnitude()
        {
            return this.GetLength();
        }

        public static double GetPolarAngle(Vector2 v)
        {
            return Math.Atan2(v.X, v.Y);
        }

        public double GetPolarAngle()
        {
            return GetPolarAngle(this);
        }

        public override bool Equals(object obj)
        {
            if (obj.GetType() != this.GetType())
            {
                return false;
            }

            return this.Equals(obj as Vector2);
        }

        public bool Equals(Vector2 obj)
        {
            if (obj is null)
            {
                return false;
            }

            if (Object.ReferenceEquals(this, obj))
            {
                return true;
            }

            return this == obj;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        public static bool operator !=(Vector2 a, Vector2 b)
        {
            return !(a == b);
        }

        public static bool operator ==(Vector2 a, Vector2 b)
        {
            return ((a.X == b.X) && (a.Y == b.Y));
        }

        public override string ToString()
        {
            return "Vector2: x = " + this.X + ", y = " + this.Y;
        }
    }
}
