﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SimpleVectorMatrixPlayground.VectorMatrix
{
    public class Vector3
    {
        public double X { private set; get; }
        public double Y { private set; get; }
        public double Z { private set; get; }

        private Vector3() { }

        private Vector3(double x, double y, double z) { this.X = x; this.Y = y; this.Z = z;  }

        public static Vector3 Create()
        {
            return new Vector3();
        }

        public static Vector3 Create(double x, double y, double z)
        {
            return new Vector3(x, y, z);
        }

        public static Vector3 Create(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            return new Vector3(x2 - x1, y2 - y1, z2 - z1);
        }

        public static Vector3 Create(Vector3 v)
        {
            return new Vector3(v.X, v.Y, v.Z);
        }

        public double GetLength()
        {
            return Math.Sqrt(X * X + Y * Y + Z * Z);
        }

        public double GetLengthSquared()
        {
            return X * X + Y * Y + Z * Z;
        }

        public static double GetDistance(Vector3 a, Vector3 b)
        {
            double x = a.X - b.X;
            double y = a.Y - b.Y;
            double z = a.Z - b.Z;
            return Math.Sqrt(x * x + y * y + z *z);
        }

        public static Vector3 operator +(Vector3 a, Vector3 b)
        {
            return Vector3.Add(a, b);
        }

        public static Vector3 Add(Vector3 a, Vector3 b)
        {
            return new Vector3(a.X + b.X, a.Y + b.Y, a.Z + b.Z);
        }

        public static Vector3 operator -(Vector3 a, Vector3 b)
        {
            return Vector3.Sub(a, b);
        }

        public static Vector3 Sub(Vector3 a, Vector3 b)
        {
            return new Vector3(a.X - b.X, a.Y - b.Y, a.Z - b.Z);
        }

        public static Vector3 operator *(Vector3 a, double b)
        {
            return Vector3.Mul(a, b);
        }

        public static Vector3 operator *(double b, Vector3 a)
        {
            return Vector3.Mul(a, b);
        }

        public static Vector3 Mul(Vector3 a, double b)
        {
            return new Vector3(a.X * b, a.Y * b, a.Z * b);
        }

        public static Vector3 operator /(Vector3 a, double b)
        {
            return Vector3.Div(a, b);
        }

        public static Vector3 Div(Vector3 a, double b)
        {
            return new Vector3(a.X / b, a.Y / b, a.Z / b);
        }

        /// <summary>
        /// Normalize a vector by dividing each component by the vector length / magnitude. The resulting vector has always length 1.
        /// </summary>
        /// <returns></returns>
        public Vector3 GetNormalized()
        {
            return Vector3.Create(this).Normalize();
        }

        public Vector3 Normalize()
        {
            double len = GetLength();
            this.X /= len;
            this.Y /= len;
            this.Z /= len;
            return this;
        }

        /// <summary>
        /// Calculates the dot product of two vectors.
        /// DotProduct(a, b) = |a| * |b| * cos(alpha).
        /// If both vectors are orthogonal, the dot product is zero.
        /// If both vectors are parallel, the dot product is the product of the length of the vectors (because cos(0) = 1).
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static double DotProduct(Vector3 a, Vector3 b)
        {
            return ((a.X * b.X) + (a.Y * b.Y) + (a.Z * b.Z));
        }

        public double DotProduct(Vector3 a)
        {
            return DotProduct(this, a);
        }

        /// <summary>
        /// Calculatres the cross product of vectors a and b. 
        /// The result vector is perpendicular to vectors a and b.
        /// Length of CrossProduct(a, b) = |a| * |b| * sin(alpha) which is the area of the parallelogram of both vectors.
        /// If both vectors are parallel, the cross product is zero.
        /// In a LHS, if b is counterclockwise from a, the crossproduct points away from the viewer.
        /// In a LHS, if b is clockwise from a, the crossproduct points towards the viewer.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static Vector3 CrossProduct(Vector3 a, Vector3 b)
        {
            double i = a.Y * b.Z - a.Z * b.Y;
            double j = a.Z * b.X - a.X * b.Z;
            double k = a.X * b.Y - a.Y * b.X;
            return Vector3.Create(i, j, k);
        }

        public Vector3 CrossProduct(Vector3 a)
        {
            return CrossProduct(this, a);
        }

        /// <summary>
        /// Returns the vector with the same length that points into the opposite direction.
        /// </summary>
        /// <returns></returns>
        public Vector3 GetNegated()
        {
            return Vector3.Create(this).Negate();
        }
        
        public Vector3 Negate()
        {
            this.X = -this.X;
            this.Y = -this.Y;
            this.Z = -this.Z;
            return this;
        }

        /// <summary>
        /// Gets the linear interpolation between two vectors.
        /// return = a + (b - a) * t.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="t"></param>
        /// <returns></returns>
        public static Vector3 GetInterpolation(Vector3 a, Vector3 b, double t)
        {
            double dx = b.X - a.X;
            double dy = b.Y - a.Y;
            double dz = b.Z - a.Z;

            return Vector3.Create(a.X + t * dx, a.Y + t * dy, a.Z + t * dz);
        }

        public Vector3 GetInterpolation(Vector3 a, double t)
        {
            return Vector3.GetInterpolation(this, a, t);
        }

        /// <summary>
        /// Gets the vector with absolute coordinate values.
        /// </summary>
        /// <returns></returns>
        public Vector3 GetAbsolute()
        {
            return Vector3.Create(this.X < 0 ? -this.X : this.X, this.Y < 0 ? -this.Y : this.Y, this.Z < 0 ? -this.Z : this.Z);
        }

        /// <summary>
        /// Returns true if a and b are orthogonal / perpendicular, otherwise false
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static bool IsOrthogonal(Vector3 a, Vector3 b)
        {
            return DotProduct(a, b) == 0;
        }

        public bool IsOrthogonal(Vector3 a)
        {
            return DotProduct(this, a) == 0;
        }


        /// <summary>
        /// Returns true if a and b are colinear, otherwise false
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static bool IsColinear(Vector3 a, Vector3 b)
        {
            return CrossProduct(a, b) == Vector3.Create();
        }

        public bool IsColinear(Vector3 a)
        {
            return CrossProduct(this, a) == Vector3.Create();
        }

        /// <summary>
        /// Gets the vector rotated by given angle in counterclockwise direction around x-axis.
        /// x coordinate remains the same.
        /// y = y * cos(a) - z * sin(a)
        /// z = y * sin(a) + z * cos(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public Vector3 GetRotatedX(double angle)
        {
            return Vector3.Create(this).RotateX(angle);
        }

        public Vector3 RotateX(double angle)
        {
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            double newY = this.Y * cos - this.Z * sin;
            double newZ = this.Y * sin + this.Z * cos;
            this.Y = newY;
            this.Z = newZ;
            return this;
        }

        /// <summary>
        /// Gets the vector rotated by given angle in counterclockwise direction around y-axis.
        /// y coordinate remains the same.
        /// x =  x * cos(a) + z * sin(a)
        /// z = -x * sin(a) + z * cos(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public Vector3 GetRotatedY(double angle)
        {
            return Vector3.Create(this).RotateY(angle);
        }

        public Vector3 RotateY(double angle)
        {
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            double newX =  this.X * cos + this.Z * sin;
            double newZ = -this.X * sin + this.Z * cos;
            this.X = newX;
            this.Z = newZ;
            return this;
        }

        /// <summary>
        /// Gets the vector rotated by given angle in counterclockwise direction around z-axis.
        /// z coordinate remains the same.
        /// x =  x * cos(a) - y * sin(a)
        /// y =  x * sin(a) + y * cos(a)
        /// </summary>
        /// <param name="angle"></param>
        /// <returns></returns>
        public Vector3 GetRotatedZ(double angle)
        {
            return Vector3.Create(this).RotateY(angle);
        }

        public Vector3 RotateZ(double angle)
        {
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            double newX = this.X * cos - this.Y * sin;
            double newY = this.X * sin + this.Y * cos;
            this.X = newX;
            this.Y = newY;
            return this;
        }

        /// <summary>
        /// Gets the non-uniform scaled vector.
        /// </summary>
        /// <param name="xs"></param>
        /// <param name="ys"></param>
        /// <param name="zs"></param>
        /// <returns></returns>
        public Vector3 GetScaled(double xs, double ys, double zs)
        {
            return Vector3.Create(this).Scale(xs, ys, zs);
        }

        public Vector3 Scale(double xs, double ys, double zs)
        {
            this.X = this.X * xs;
            this.Y = this.Y * ys;
            this.Z = this.Z * zs;
            return this;
        }

        /// <summary>
        /// Get the length of the vector when Vector b is projected onto vector a.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static double GetProjectedVectorLength(Vector3 a, Vector3 b)
        {
            return DotProduct(a, b) / a.GetLength();
        }

        /// <summary>
        /// Get the length of the vector when Vector v is projected onto current vector.
        /// </summary>
        /// <param name="a"></param>
        /// <returns></returns>
        public double GetProjectedVectorLength(Vector3 v)
        {
            return DotProduct(this, v) / v.GetLength();
        }

        /// <summary>
        /// Get the result vector when Vector b is projected onto vector a.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static Vector3 GetProjectedVector(Vector3 a, Vector3 b)
        {
            return a.GetNormalized() * DotProduct(a, b) / a.GetLength();
        }

        /// <summary>
        /// Get the result vector when Vector v is projected onto current vector.
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public Vector3 GetProjectedVector(Vector3 v)
        {
            return this.GetNormalized() * DotProduct(this, v) / this.GetLength();
        }

        /// <summary>
        /// Get the reflected vector a at the surface given with normal n.
        /// </summary>
        /// <param name="a">The vector a to reflect at the surface</param>
        /// <param name="n">Normal of the surface. Bit be normalited (length = 1) of course-</param>
        /// <returns></returns>
        public static Vector3 GetReflectedVector(Vector3 a, Vector3 n)
        {
            Vector3 r = a - (2 * DotProduct(a, n) * n);
            return r;
        }

        /// <summary>
        /// Get the cartesian coordinates from the spherical coordinates.
        /// x = r * Sin(a1) * Cos(a2)
        /// y = r * Sin(a1) * Sin(a2)
        /// z = r * Cos(a1)
        /// </summary>
        /// <param name="mag">radial distance</param>
        /// <param name="angle1">polar angle (in x-z plane)</param>
        /// <param name="angle2">azimuthal angle (in x-y plane)</param>
        /// <returns></returns>
        public static Vector3 GetVectorFromSpherical(double mag, double angle1, double angle2)
        {
            double x = mag * Math.Sin(angle1) * Math.Cos(angle2);
            double y = mag * Math.Sin(angle1) * Math.Sin(angle2);
            double z = mag * Math.Cos(angle1);
            return Vector3.Create(x, y, z);
        }

        public double GetSphericalMagnitude()
        {
            return this.GetLength();
        }

        public static double GetPolarAngle(Vector3 v)
        {
            int factor = (v.X < 0) ? -1 : 1;
            return factor * Math.Acos(v.Z / v.GetLength());
        }

        public double GetPolarAngle()
        {
            return GetPolarAngle(this);
        }

        public static double GetAzimuthalAngle(Vector3 v)
        {
            return Math.Atan2(v.X, v.Y);
        }

        public double GetAzimuthalAngle()
        {
            return GetAzimuthalAngle(this);
        }

        public override bool Equals(object obj)
        {
            if (obj.GetType() != this.GetType())
            {
                return false;
            }

            return this.Equals(obj as Vector3);
        }

        public bool Equals(Vector3 obj)
        {
            if (obj is null)
            {
                return false;
            }

            if (Object.ReferenceEquals(this, obj))
            {
                return true;
            }

            return this == obj;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        public static bool operator !=(Vector3 a, Vector3 b)
        {
            return !(a == b);
        }

        public static bool operator ==(Vector3 a, Vector3 b)
        {
            return ((a.X == b.X) && (a.Y == b.Y) && (a.Z == b.Z));
        }

        public override string ToString()
        {
            return "Vector3: x = " + this.X + ", y = " + this.Y + ", z = " + this.Z;
        }
    }
}
